module Dijkstra (
                 --testshortestpath,
                 best_step,distanceToPoints,
--                 distanceToPointsSimple,
                 --distanceToPointsLimited,
                        -- the following exported to keep -Wall happy
                 --SearchType (..),emptyboard,myboard,show_all,
                 DijkstraHeading,next_step
                ) where

import Maybe
import Data.FiniteMap
import Data.Set
import Board
import Types
import Data.List
-- import System.CPUTime

-- import Debug.Trace

data SearchType = FixedStart | FixedEnd deriving (Show,Eq)

newtype SearchResult a n = SR (FiniteMap a (n,Path a),FiniteMap a (n,Path a))

instance Show (SearchResult a n) where
   show _ = "<search result>"

--show_all :: (Show a,Ord a,Show n) => SearchResult a n -> String
--show_all (SR (done,nearlydone)) = "\ndone:\n" ++ showFM done 
--                               ++ "\nnearlydone:\n"++ showFM nearlydone ++ "\n"

--showFM :: (Show a,Ord a,Show b) => FiniteMap a b -> String
--showFM fm = concat $ map showItem $ fmToList fm
--        where showItem (a,b) = "  " ++ show a ++ " -> " ++ show b ++ "\n"

findres :: Ord a => SearchType -> SearchResult a n -> a -> Maybe (n,Path a)
findres searchtype (SR (done,_)) target
  = fmap fixup $ lookupFM done target
         where fixup (dist,path) | searchtype == FixedStart 
                                     = (dist,tail $ reverse (target:path))
                                 | otherwise
                                     = (dist,path)

initres :: (Ord a,Num n) => a -> SearchResult a n
initres source = SR (emptyFM,unitFM source (0,[]))

{-
sizeres :: Ord a => SearchResult a n -> Int
sizeres (SR (done,nearlydone)) = sizeFM done + sizeFM nearlydone
-}

data SearchLimit n = NoLimit | Limit n deriving Eq
instance Functor SearchLimit where
   fmap _f NoLimit = NoLimit
   fmap f (Limit n) = Limit (f n)

shortestpath ::
-- a is the type of coordinates, n the type of distances
   (Ord a,Num n,Ord n,Show a) =>
-- adjacency function, takes a bias factor, a vertex 
-- and returns a list of adjacent vertices together with biased distances
   ((a -> n) -> a -> [(a,n)]) ->
-- bias factor calculation; given two positions, this will calculate how
-- much our distance function will be skewing the results
   (a -> n) -> 
-- target
   a ->
-- limit on the iterations, 
-- and result from a previous search with the same source
   (SearchLimit Int,SearchResult a n) ->
-- remaining iterations and result
   (SearchLimit Int,SearchResult a n)

shortestpath adj biasfunc target (itlimit,searchres)
 = 
--   (\sr -> trace (show sr) sr) $ 
   apsnd unbiassr $ 
--     (\sr -> trace (show sr) sr) $ 
     step itlimit $ 
       biassr searchres
     where 

           bias = mapFM (\pos (dist,path) -> (dist+biasfunc pos,path))
           unbias = mapFM (\pos (dist,path) -> (dist-biasfunc pos,path))

           apsnd f (a,b) = (a,f b)

           biassr (SR (done,nearlydone)) = SR (bias done,bias nearlydone)
           unbiassr (SR (done,nearlydone)) = SR (unbias done,unbias nearlydone)

           step limit sr@(SR (done,nearlydone))
             | limit == Limit 0 = (limit,sr)
             | isEmptyFM nearlydone = (limit,sr)
             | newpos == target = (limit,newsr)
             | otherwise = -- trace ("visiting " ++ show newpos) $
                           step (fmap (\x -> x-1) limit) newsr
            where 
                  newsr = SR (addToFM done newpos newdetails,
                              delFromFM nearlydone' newpos)

                  (newpos,newdetails@(newdist,newpath)) 
                     = fromJust $ findmin nearlydone

                  findmin gm = foldFM getmin Nothing gm
                  getmin pos (dist,path) Nothing = Just (pos,(dist,path))
                  getmin pos (dist,path) (Just (pos',(dist',path')))
                     = if dist<dist' then Just (pos,(dist,path)) 
                                     else Just (pos',(dist',path'))

                  nearlydone' = foldr addnd nearlydone (adj biasfunc newpos)
                  addnd (pos',dist') nd 
                    = case lookupFM done pos' of
                       Just _ -> nd
                       Nothing 
                         -> addToFM_C chooseBest nd pos' 
                                      (newdist+dist',newpos:newpath)

                  chooseBest (dist1,path1) (dist2,path2)
                      = if dist1<dist2 then (dist1,path1) else (dist2,path2)

{-
normdist :: Position -> Position -> Distance
normdist (x,y) (x',y') = floor (sqrt (sqdiff x x' + sqdiff y y') :: Float)
                             where sqdiff a a' = fromIntegral (a-a') ** 2
-}

meshdist :: Position -> Position -> Distance
meshdist (x,y) (x',y') = abs (x-x') + abs (y-y')

{-
sqnormdist :: Position -> Position -> Distance
sqnormdist (x,y) (x',y') = sq (x-x') + sq (y-y') where sq a = a*a

minlinedist :: Coord -> Coord -> Position -> Distance
minlinedist xline yline (x',y') = min (abs (x'-xline)) (abs (y'-yline))

pairdist :: (Coord,Coord) -> Coord -> Distance
pairdist (l,r) p = if p<=l then l-p else if r<=p then p-r else 0
-}

mydist :: Position -> Position -> Position -> Distance
mydist (_,_) (targetx,targety) (x,y) =
   99 * meshdist (targetx,targety) (x,y)

{-
 = if x<=lx then lx-x + pairdist (by,ty) y
   else if y<=by then by-y + pairdist (lx,rx) y
   else (x-lx) + (y-by)
     where (lx,rx) = if sourcex<targetx then (sourcex,targetx) 
                                        else (targetx,sourcex)
           (by,ty) = if sourcey<targety then (sourcey,targety)
                                        else (targety,sourcey)
-}


boardadj :: Board -> Set Position ->
                (Position -> Distance) -> Position -> [(Position,Distance)]
boardadj board fps = boardadj'
        where 
              -- might add precalculation of neighbours here
              boardadj' weight (x,y) =
               map retval $ reachable_neighbors board (x,y)
                  where 
                        xyweight = weight (x,y)
                        origininset = elementOf (x,y) fps
                        dist (x',y') = if origininset || elementOf (x',y') fps
                                       then 99 else 100
                        weightdist (x',y') 
                            = dist (x',y') + weight (x',y') - xyweight
                        retval pos = (pos,weightdist pos)

--emptyboard :: Board
--emptyboard = make_board 1000 1000 (repeat ".")

{-
testshortestpath :: Board -> [Position] -> SearchType -> Position -> Position 
                               -> (Int,Maybe (Distance,Path Position))

testshortestpath board fps searchtype start end = -- trace (show sr) $
    (sizeres sr,
     findres searchtype sr (if searchtype == FixedStart then end else start))
      where 
            (_,sr) = shortestpath (boardadj board (mkSet fps)) 
                                  (mydist source target)
                                  target
                                  (NoLimit,initres source)

            (source,target) = if searchtype == FixedStart then (start,end)
                                                          else (end,start)
-}

type DijkstraHeading = (Position -- target
                       ,Path Position -- last chosen path
                       ,Maybe (SearchResult Position Distance) -- cached info
                       )




best_step :: Board -> [Position] -> Maybe DijkstraHeading
                -> Position -> [Position] -> Maybe (Direction,DijkstraHeading)
best_step board fps dijhead curpos wantedposs =
  case bestinfo of
     Nothing -> Nothing
     Just (_,[]) -> Nothing
     Just (_,path@(nextpos:rempath)) 
       -> Just (dir,(target,rempath,searchres))
            where dir = fromJust $ calc_direction curpos nextpos
                  target = last path
                  searchres 
                    = case dijhead of
                        Just (target',_,Just sr) | target == target'
                          -> Just sr
                        _ -> Nothing
    where

          bestinfo = foldr chooseinfo Nothing info
          chooseinfo Nothing Nothing = Nothing
          chooseinfo (Just i) Nothing = Just i
          chooseinfo Nothing (Just i) = Just i
          chooseinfo (Just (d,p)) (Just (d',p')) = if d<d' then Just (d,p)
                                                           else Just (d',p')

          info = map getinfo wantedposs

          wantedposs' = sortByMetric (meshdist curpos) wantedposs
          fullsearch  = resof $
                         foldl searchone' 
                               (Left (maxBound,NoLimit,initres curpos)) $
                            zip wantedposs' 
                                (tail (map Just wantedposs' ++ [Nothing]))

          resof (Left (_,_,sr)) = sr
          resof (Right sr) = sr

          searchone' (Right sr) _ = Right sr
          searchone' (Left (mindist,l,sr)) (wpos,mwpos')
            = let (l',sr') = searchone wpos (l,sr)
                  mindist' = case findres FixedStart sr' wpos of
                              Just (dist,_) -> min mindist dist
                              _ -> mindist
              in case mwpos' of
                   Just wpos' | mindist' <= 99*meshdist curpos wpos' 
                     -> Right sr'
                   _ -> Left (mindist',l',sr')

          searchone wantedpos lsr 
             = shortestpath (boardadj board (mkSet fps))
                            (mydist curpos wantedpos)
                            wantedpos
                            lsr
          getinfo wantedpos = findres FixedStart fullsearch wantedpos


next_step :: Board -> [Position] -> Position -> DijkstraHeading 
                        -> Maybe (Direction,DijkstraHeading)
next_step board fps curpos (target,path,searchresm) =
  case path of
    (nextpos:path') -> case calc_direction curpos nextpos of
                         Just d -> Just (d,(target,path',searchresm))
                         Nothing -> searchagain
    [] -> searchagain
   where searchagain 
           = case findres FixedEnd searchres' curpos of
               Just (_,nextpos:path') 
                 -> Just (fromJust $ calc_direction curpos nextpos,
                          (target,path',Just searchres')
                         )
               _ -> Nothing

         (_,searchres') = shortestpath (boardadj board (mkSet fps))
                                       (mydist target curpos)
                                       curpos
                                       (NoLimit,searchres)

         searchres = fromMaybe (initres target) searchresm

distanceToPoints :: Board -> [Position] 
                        -> Position -> [Position] -> [Maybe Distance]
distanceToPoints board fps curpos wantedposs =
  map (fmap fst) (info_list board fps NoLimit curpos wantedposs)

{-
distanceToPointsSimple :: Board -> Position -> [Position] -> [Maybe Distance]
distanceToPointsSimple board curpos wantedposs =
  map (fmap (`div` 100)) $ distanceToPoints board [] curpos wantedposs
-}

{-
distanceToPointsLimited :: Board -> Distance -> Position -> [Position] 
                               -> [Maybe Distance]
distanceToPointsLimited board radius curpos wantedposs = 
  map (fmap ((`div` 100) . fst) . getinfo) wantedposs 
   where (_,fullsearch) = foldr searchone 
                                (Limit its,initres curpos) 
                                wantedposs
         its = sq radius + sq (radius+1)
         sq x = x*x
         searchone wantedpos lsr
             = shortestpath (boardadj board emptySet)
                            (const 0) -- no biasing
                            wantedpos
                            lsr
         getinfo wantedpos = findres FixedStart fullsearch wantedpos
-}

sortByMetric :: Ord metric => (a -> metric) -> [a] -> [a]
sortByMetric f xs = sortBy f' xs
  where f' a b = compare (f a) (f b)

info_list :: Board -> [Position] -> SearchLimit Int
                   -> Position -> [Position]
                   -> [Maybe (Distance,Path Position)]
info_list board fps sl curpos wantedposs = -- trace (show fullsearch) $ 
  map getinfo wantedposs
    where 
          wantedposs' = sortByMetric (meshdist curpos) wantedposs
          (_,fullsearch) = foldr searchone (sl,initres curpos) wantedposs'
          searchone wantedpos lsr 
             = shortestpath (boardadj board (mkSet fps))
                            (mydist curpos wantedpos)
                            wantedpos
                            lsr
          getinfo wantedpos = findres FixedStart fullsearch wantedpos


