
\begin{code}
module Main (main) where

import Monad
import IO
import Data.IORef (IORef, newIORef, readIORef, writeIORef)
import System (getArgs, {-exitWith, ExitCode(ExitSuccess)-})
-- import Socket (withSocketsDo)

import TCP (connect, initialise, turn_info, command)
import Knowledge
import AI (think)
import Board
-- import Adrenaline
import Types

import GHC.Exception
import System.Random

main :: IO ()
-- main = withSocketsDo $ catch realmain (\_ -> exitWith ExitSuccess)
main = realmain

realmain :: IO ()
realmain = do 
       hSetBuffering stdout LineBuffering
       (host:port:rest) <- getArgs
       server <- connect host port
       let debug = case rest of
                     "debug":_ -> True
                     _ -> False
       (b, configuration) <- initialise server
--       print board
       when debug $ print configuration
       kref <- newIORef (init_knowledge b configuration)
       sequence_ $ repeat $ catcherrors kref server $
           do
           info@(_, packages) <- turn_info debug server

           k_old <- readIORef kref
           let k = update_knowledge info k_old

           when (debug && backing_off k) $
                putStrLn ("** Back off! **")

           {-
           when (on_adrenaline k)
                putStrLn "*** ADRENALINE ***"
           -}

           let (bid, cmd, k') = think k
                                
           when debug $ dump_log cmd k info-- dump_log Command Knowledge info
           send_command k server bid cmd
           writeIORef kref (remember_bid_cmd bid cmd k')

send_command :: Knowledge -> Handle -> Bid -> Command -> IO ()
send_command k server bid cmd
    = case cmd of
          Dunno -> do c <- case reachable_directions (board k) (my_pos k) of
                              [] -> return $ Pick []
                              ds -> do d <- pickRandomElem ds
                                       return $ Move d
                      command server 1 c
          _ -> command server bid cmd

pickRandomElem :: [a] -> IO a
pickRandomElem [] = error "competely fatal"
pickRandomElem ls = do gen <- newStdGen
                       let (n,_) = next gen
                       return $ ls!!(n `mod` length ls)

catcherrors :: IORef Knowledge -> Handle -> IO () -> IO ()
catcherrors kref server loop 
  = catchException 
       loop
       (\e -> case e of
               IOException _ -> ioError e
               _ -> 
                do putStr "would have died: "
                   print e
                   k <- readIORef kref
                   send_command k server 1 Dunno
       )


ctch :: String -> a -> IO a -> IO a
ctch str def comp 
  = catchException comp (\e -> (putStr str >> print e >> return def))

-- unified debugging at some point?
dump_log       :: (Show a) => Command -> Knowledge -> a -> IO ()
dump_log c k i = do h <- openFile ("board-" ++ show (name k)) AppendMode
                    ctch "ignore:" () $ hPutStrLn h ("\nbegin debug, Knowledge:\n" ++
                                                     show k ++ 
                                                     "\nresulting move:\n" ++ 
                                                     show c ++
                                                     "\ninfo:\n" ++
                                                     show i ++
                                                     "show_board:\n" ++
                                                     (show_board b mp rs))
                    hClose h
    where b = board k
          mp = my_pos k
          rs = map bot_pos $ others k

\end{code}
