
\begin{code}
module TCP (connect, initialise, turn_info, command) where

import Socket
import IO
import Types
import Board
import Monad

connect :: String -> String -> IO Handle
connect hostname port = do a <- connectTo hostname (PortNumber portnum)
                           hSetBuffering a NoBuffering
                           return a
    where portnum = fromIntegral (read port :: Integer)

break_space :: String -> (String, String)
break_space s = case break (' ' ==) s of
                    (x, y) -> (x, drop 1 y)

break_spaces :: String -> [String]
break_spaces xs = case break_space xs of
                      (_, "") -> [xs]
                      (ys, zs) -> ys:break_spaces zs

initialise :: Handle -> IO (Board, Configuration)
initialise a = do hPutStrLn a "Player"
                  dimensions <- hGetLine a
                  let (d1, d2) = break_space dimensions
                      width = read d1
                      height = read d2
                  board_lines <- read_board a height
                  let board = make_board width height board_lines
                  config <- read_configuration a
                  return (board, config)

turn_info :: Bool -> Handle -> IO Turn_Info
turn_info debug a = do wh <- hGetLine a
                       ps <- hGetLine a
                       when debug $ print (wh, ps)
                       return (parse_happenings (-1) wh, parse_pkgs ps)

parse_happenings :: ID -> String -> [Update]
parse_happenings _ "" = []
parse_happenings i s = case head this of
                           '#' -> let i' = read (tail this)
                                  in (i', Alive):parse_happenings i' s'
                           'N' -> (i, HN):parse_happenings i s'
                           'S' -> (i, HS):parse_happenings i s'
                           'E' -> (i, HE):parse_happenings i s'
                           'W' -> (i, HW):parse_happenings i s'
                           'P' -> let (n, s'') = break_space s'
                                  in (i, HP (read n)):parse_happenings i s''
                           'D' -> let (n, s'') = break_space s'
                                  in (i, HD (read n)):parse_happenings i s''
                           'X' -> let (x, s1) = break_space s'
                                      (_, s2) = break_space s1 -- 'Y'
                                      (y, s3) = break_space s2
                                      pos = (read x, read y)
                                  in (i, HPos pos):parse_happenings i s3
                           _ -> parse_happenings i s' -- error "parse_happenings"
    where (this, s') = break_space s

parse_pkgs :: String -> [Package]
parse_pkgs "" = []
parse_pkgs xs = parse (break_spaces xs)
  where
  parse [] = []
  parse (i:x:y:w:rest) = ((read i), (read x, read y), (read w))
                         : parse rest
  parse _ = [] -- error "Couldn't parse package info"

command :: Handle -> Bid -> Command -> IO ()
command a b c = do hPutStr a (show b)
                   hPutChar a ' '
                   hPutStrLn a (show_command c)

read_board :: Handle -> Coord -> IO [String]
read_board _ 0 = return []
read_board a (n+1) = do line <- hGetLine a
                        rest <- read_board a n
                        return (line:rest)

read_configuration :: Handle -> IO Configuration
read_configuration a = do line <- hGetLine a
                          let (l1, l') = break_space line
                              (l2, l3) = break_space l'
                          return (read l1, read l2, read l3)
\end{code}

