
\begin{code}
module Types where

type Count = Int
type ID = Int
type Bid = Int
type Weight = Int
type Money = Int
type Coord = Int
type Position = (Coord, Coord)
type Configuration = (ID, Weight, Money)
-- type Robot = (ID, Position, Count)
data Robot = Robot { bot_id :: ID,
                     bot_pos :: Position,
                     bot_count :: Count} deriving Show

type Path a = [a]
type Distance = Int

type Update    = (ID, Happened)

type Package   = (ID, Position, Weight) -- Position is destination

package_id :: Package -> ID
package_id (i, _, _) = i
package_pos :: Package -> Position
package_pos (_, p, _) = p
package_weight :: Package -> Weight
package_weight (_, _, w) = w

pkg_id :: Package -> ID
pkg_id (j,_,_) = j


type Turn_Info = ([Update], [Package])


data Happened = HN | HS | HW | HE | HPos Position | HP ID | HD ID | Alive
    deriving (Show, Eq)

data Command = Move Direction | Pick [ID] | Drop [ID] | Dunno
    deriving Show

data Direction = N | S | E | W
    deriving Show

show_command :: Command -> String
show_command (Move d) = "Move " ++ show d
show_command (Pick ids) = "Pick" ++ show_ids ids
show_command (Drop ids) = "Drop" ++ show_ids ids
show_command Dunno = "Pick" -- Wrong, but it might get sent to the server...

show_ids :: [ID] -> String
show_ids = concat . map ((' ':) . show)

data Terrain = Free | Water | Wall | Base
  deriving (Eq, Show)

-- instance Show Terrain where
--     showsPrec d t =
--         showParen (d >= 10) $ showChar $ show_chr t
--       where
--       show_chr :: Terrain -> Char
--       show_chr Free = '.'
--       show_chr Water = '~'
--       show_chr Wall = '#'
--       show_chr Base = '@'

instance Read Terrain where
    readsPrec _ = readsTerrain
      where
      readsTerrain :: ReadS Terrain
      readsTerrain (c:cs) =
          case char_to_terrain c of
               Nothing -> []
               Just t -> [(t,cs)]
      readsTerrain [] = []
      char_to_terrain :: Char -> Maybe Terrain
      char_to_terrain '.' = Just Free
      char_to_terrain '~' = Just Water
      char_to_terrain '#' = Just Wall
      char_to_terrain '@' = Just Base
      char_to_terrain _ = Nothing

calc_direction :: Position -> Position -> Maybe Direction
calc_direction (x,y) (x',y') | x'-x == 1 && y'==y = Just E
                             | x-x' == 1 && y'==y = Just W
                             | y'-y == 1 && x'==x = Just N
                             | y-y' == 1 && x'==x = Just S
                             | otherwise = Nothing

\end{code}

